"""
Module data_comparator.

Ce module contient la classe DataComparator qui permet de comparer les données
entre deux sources SQL en utilisant des DataFrames Polars, avec support du
traitement par lots et de la vérification des types de colonnes.
"""
from collections.abc import Generator

import polars as pl
from polars.testing import assert_frame_equal

from .data_factory import DataFactory
from .logger import Logger
from .timer import timer


class DataComparator:
    """
    Classe pour comparer les données provenant de deux sources SQL différentes.

    Cette classe permet de récupérer les données de deux bases de données via des requêtes
    SQL, de les transformer en DataFrames Polars et de comparer les résultats par lots
    (batchs). Elle prend en charge la vérification des valeurs et, optionnellement, des types
    de colonnes.
    """

    def __init__(self, dictionnary: dict, logger: Logger) -> None:
        """
        Initialise un dataComparator avec les informations de connexions aux bases de données.

        :param dictionnary: le dictionnaire qui contient les informations du pipeline
            - 'db_source_1': la base de données source 1
            - 'query_source_1': la requête à envoyer à la source 1
            - 'db_source_2': la base de données source 2
            - 'query_source_2': la requête à envoyer à la source 2
            - 'batch_size': la taille des lots pour le traitement en batch
        :param logger: le logger pour gérer la journalisation des évènements du pipeline
        """
        self.logger = logger
        self.__db_source_1 = dictionnary.get("db_source_1")
        self.__query_source_1 = dictionnary.get("query_source_1")
        self.__db_source_2 = dictionnary.get("db_source_2")
        self.__query_source_2 = dictionnary.get("query_source_2")
        self.__batch_size = dictionnary.get("batch_size", 10_000)

    def _fetch_data(self, db: DataFactory, query: str) -> Generator[pl.DataFrame]:
        """
        Exécute la requête donnée sur la base spécifiée et retourne un DataFrame Polars.

        :param db: la base sur laquelle envoyer la requête
        :param query: la requête à envoyer
        :return: le résultat de la requête sous forme de DataFrame
        """
        db.connect()
        for batch in db.sql_query(query):
            yield pl.DataFrame(batch, orient="row", strict=False, infer_schema_length=self.__batch_size)

    @timer
    def compare(self, *, check_dtypes: bool = False) -> bool:
        """
        Compare les résultats des deux requêtes SQL entre les deux bases de données.

        :param check_dtypes: si True, vérifie aussi la stricte égalité des types de colonnes
                             si False (défaut), ne vérifie que les valeurs
        :return: True si les DataFrames sont identiques (selon le niveau de vérification)
        """
        try:
            gen1 = self._fetch_data(self.__db_source_1, self.__query_source_1)
            gen2 = self._fetch_data(self.__db_source_2, self.__query_source_2)
            batch_index = 0
            while True:
                try:
                    df1 = next(gen1)
                except StopIteration:
                    df1 = None
                try:
                    df2 = next(gen2)
                except StopIteration:
                    df2 = None
                if df1 is None and df2 is None:
                    self.logger.info("Les données des deux sources sont identiques.")
                    return True
                if (df1 is None) != (df2 is None):
                    self.logger.error("Nombre de lignes différent entre les deux sources.")
                    return False
                batch_index += 1
                try:
                    assert_frame_equal(df1, df2, check_dtypes=check_dtypes)
                except AssertionError as e:
                    self.logger.error(f"Les données diffèrent dans le batch #{batch_index}: {e}")
                    return False
                self.logger.debug(f"Batch #{batch_index} identique ({len(df1)} lignes).")
        except Exception as e:
            self.logger.error(f"Erreur lors de la comparaison des données : {e}")
            raise



"""
Module mail_reader.

Contient la classe MailReader pour se connecter à un serveur IMAP et lire, filtrer et gérer les emails.
"""
from __future__ import annotations

import email
import imaplib
import mimetypes
import re
from email import policy
from http import HTTPStatus
from typing import TYPE_CHECKING

import chardet
import requests

if TYPE_CHECKING:
    from mailbox import Message

    from .logger import Logger

class MailReader:
    """Classe pour se connecter à un serveur IMAP et lire, filtrer et gérer les emails."""

    def __init__(self, configuration: dict, logger: Logger, auth_mode: str) -> None:
        """
        Instancie la classe MailReader.

        :param configuration: Un dictionnaire contenant tous les paramètres nécéssaires pour lancer une connexion imap
        :param logger: Un logger ads qui va gérer les logs des actions de la classe
        :param auth_mode: Le mode d'authentification 'oauth2' ou 'credentials'
        """
        self.logger = logger
        self.__url=configuration.get("url")
        self.__clientId=configuration.get("clientId")
        self.__clientSecret=configuration.get("clientSecret")
        self.__email=configuration.get("email")
        self.__emailFrom=configuration.get("emailFrom")
        self.__subject=configuration.get("subject")
        self.__scope=configuration.get("scope")
        self.__server=configuration.get("server")
        self.__login=configuration.get("login")
        self.__password=configuration.get("password")
        self.__token=None
        self.__connection=None
        if auth_mode == "oauth2":
            self.__connect_with_token()
        elif auth_mode == "credentials":
            self.__connect_with_credentials()
        else:
            msg = f"Le mode d'authentification {auth_mode} n'est pas reconnu. Utiliser 'oauth2' ou 'credentials'"
            self.logger.error(msg)
            raise ValueError(msg)

    def __connect_with_credentials(self) -> None:
        """Se connecte à un server imap et récupère le token."""
        self.logger.info("Tentative de connexion avec le serveur mail.")
        try:
            imap_conn = imaplib.IMAP4_SSL(self.__server)
            imap_conn.login(self.__login, self.__password)
            self.__connection = imap_conn
            self.logger.info("Connexion établie.")
        except Exception as e:
            self.logger.error(f"La tentative de connexion a échoué: {e!s}")
            raise

    def __connect_with_token(self) -> None:
        """Se connecte à un server imap et récupère le token."""
        self.logger.info("Tentative de connexion avec le server mail.")
        payload = {
            "grant_type": "client_credentials",
            "client_id": self.__clientId,
            "client_secret": self.__clientSecret,
            "scope": self.__scope,
        }
        response = requests.post(self.__url, data=payload, timeout=(5, 15))
        if response.status_code == HTTPStatus.OK:
            self.logger.info("Connexion établie")
            self.logger.debug("Tentative de récupération du token.")
            token_data = response.json()
            if "access_token" in token_data:
                self.logger.debug("Token récupéré avec succès !")
                self.__token = token_data["access_token"]
                imap_conn = imaplib.IMAP4_SSL(self.__server)
                auth_string = f"user={self.__email}\1auth=Bearer {self.__token}\1\1"
                imap_conn.authenticate("XOAUTH2", lambda _: auth_string.encode("utf-8"))
                self.__connection = imap_conn
            else:
                msg = f"Le token n'a pas pu être récupéré {response.reason}."
                self.logger.error(msg)
                raise RuntimeError(msg)

        else:
            msg = f"La tentative de connexion a échoué: {response.reason}"
            self.logger.error(msg)
            raise RuntimeError(msg)

    def mark_as_seen(self, e_id: str) -> None:
        """Marque le mail identifié comme lu."""
        imap_conn: imaplib.IMAP4_SSL = self.__connection
        self.logger.info(f"Changement du flag du mail {e_id} à 'Seen'.")
        try:
            imap_conn.store(e_id, "+FLAGS", "\\Seen")
        except Exception as e:
            self.logger.error(f"Le changement de flag vers 'Seen' du mail {e_id} a échoué: {e!s}.")
            raise

    def mark_as_un_seen(self, e_id: str) -> None:
        """Marque le mail identifié comme non-lu."""
        imap_conn: imaplib.IMAP4_SSL = self.__connection
        self.logger.info(f"Changement du flag du mail {e_id} à 'UnSeen'.")
        try:
            imap_conn.store(e_id, "-FLAGS", "\\Seen")
        except Exception as e:
            self.logger.error(f"Le changement de flag vers 'UnSeen' du mail {e_id} a échoué: {e!s}.")
            raise

    def delete(self, e_id: str) -> None:
        """Supprime l'email correspondant à l'identifiant."""
        imap_conn: imaplib.IMAP4_SSL = self.__connection
        self.logger.info(f"Suppression du mail {e_id}.")
        try:
            imap_conn.store(e_id, "+FLAGS", "\\Deleted")
        except Exception as e:
            self.logger.error(f"La suppression du mail {e_id} a échoué: {e!s}.")
            raise

    def _validate_flag(self, flag: str) -> None:
        valid_flags = {
            "ALL", "SEEN", "UNSEEN", "FLAGGED", "UNFLAGGED", "ANSWERED", "UNANSWERED", "DELETED", "UNDELETED", "DRAFT",
        }
        if flag.upper() not in valid_flags:
            msg = f"Flag incorrect: {sorted(valid_flags)}"
            self.logger.error(msg)
            raise ValueError(msg)

    def _search_emails(
            self,
            imap_conn: imaplib.IMAP4_SSL,
            flag: str,
            email_from: str | None,
            charset: str | None,
    ) -> tuple[str, list[str]]:
        self.logger.info("Tentative de récupération des mails.")
        if email_from:
            return imap_conn.search(charset, f'{flag} FROM "{email_from}"')
        return imap_conn.search(charset, flag)

    def _filter_by_subject(self, imap_conn: imaplib.IMAP4_SSL, email_ids: list[str], email_subject: str) -> list[str]:
        matching_emails: list[str] = []
        for e_id in email_ids:
            status, msg_data = imap_conn.fetch(e_id, "(RFC822)")
            if status != "OK":
                msg = f"Erreur lors du filtrage par sujet, {status, email_ids[0]}"
                self.logger.error(msg)
                raise RuntimeError(msg)
            for response_part in msg_data:
                if isinstance(response_part, tuple):
                    msg = email.message_from_bytes(response_part[1])
                    subject = msg.get("Subject", "")
                    if re.search(email_subject, subject, re.IGNORECASE):
                        matching_emails.append(e_id)
        return matching_emails

    def get_email(
        self,
        mailbox: str = "INBOX",
        email_from: str | None = None,
        email_subject: str | None = None,
        flag: str = "UnSeen",
        charset: str | None = None,
        *,
        readonly: bool = False,
    ) -> list[str]:
        """
        Récupère les mails selon les paramètres renseignés.

        :param mailbox: Dossier IMAP à lire (ex: INBOX)
        :param email_from: Filtrer sur l'expéditeur
        :param email_subject: Filtrer sur le sujet
        :param flag: Critère IMAP (ALL, SEEN, UNSEEN, etc.)
        :param charset: Jeu de caractères pour la recherche (ex: UTF-8)
        :param readonly: Si False, les mails seront marqués comme lus
        :return: Liste des identifiants d'emails correspondants
        """
        self._validate_flag(flag)
        imap_conn = self.__connection
        imap_conn.select(mailbox.upper(), readonly=readonly)
        status, email_ids = self._search_emails(imap_conn, flag, email_from, charset)
        if status != "OK":
            msg = f"Erreur lors de la récupération des emails, {status, email_ids[0]}"
            self.logger.error(msg)
            raise RuntimeError(msg)
        if not email_ids[0]:
            self.logger.warning("Aucun email n'a été trouvé.")
            return []
        email_ids = email_ids[0].split()
        if email_subject:
            matching_emails = self._filter_by_subject(imap_conn, email_ids, email_subject)
        else:
            matching_emails = email_ids
        self.logger.info(f"{len(email_ids)} mail(s) trouvé(s)")
        return matching_emails

    def _decode_email_body(self, msg: Message) -> str:
        """Extrait et décode le texte de l'email."""
        body = ""
        for part in msg.walk():
            if part.get_content_type() in ["text/plain", "text/html"]:
                raw = part.get_payload(decode=True)
                encoding = chardet.detect(raw)["encoding"] or part.get_content_charset() or "utf-8"
                self.logger.debug(f"Encodage détecté: {encoding} pour {part.get_content_type()}")
                try:
                    body = raw.decode(encoding, errors="replace")
                except LookupError:
                    body = raw.decode("iso-8859-1", errors="replace")
        return body

    def __get_attachment(self, msg: Message) -> list[dict]:
        """Récupère les pièces jointes d'un email et renvoie une liste de dictionnaires."""
        def _decode_text(raw: bytes, filename: str) -> tuple[str, str]:
            """Décode le contenu texte et renvoie (content, encoding)."""
            encoding = chardet.detect(raw)["encoding"] or "uft-8"
            try:
                return raw.decode(encoding), encoding
            except UnicodeDecodeError:
                self.logger.error(f"Échec du décodage avec {encoding} pour {filename}. Enregistrement en binaire.")
                return raw.decode("iso-8859-1", errors="replace"), "iso-8859-1"

        def _process_part(part: Message) -> dict | None:
            if part.get_content_disposition() != "attachment":
                return None
            filename = part.get_filename()
            if not filename:
                return None
            self.logger.info(f"Téléchargement de la pièce jointe: {filename}.")
            raw = part.get_payload(decode=True)
            mime_type, _ = mimetypes.guess_type(filename)
            is_text = mime_type and mime_type.startswith("text")
            content: bytes | str = raw
            encoding: str | None = None
            binary = not is_text
            if is_text:
                content, encoding = _decode_text(raw, filename)
                binary = False
            return {
                "filename": filename,
                "content": content,
                "binary": binary,
                "encoding": encoding if not binary else "binary",
            }
        attachments = [att for part in msg.walk() if (att := _process_part(part))]
        if self.logger and not attachments:
            self.logger.info("Aucune pièce jointe n'a été trouvée.")
        return attachments

    def read_email(self, email_id: str, *, get_attchment: bool=False) -> dict:
        """Lit un email à partir de son ID et retourne son contenu."""
        self.logger.info(f"Tentative de lecture de l'email {email_id}.")
        imap_conn = self.__connection
        status, msg_data = imap_conn.fetch(email_id, "(RFC822)")
        if status != "OK":
            msg = f"Erreur lors de la lecture de l'email {email_id}."
            self.logger.error(msg)
            raise RuntimeError(msg)
        if not msg_data[0]:
            self.logger.warning(f"Le contenu du mail {email_id} est vide.")
            return {"message": ""}
        raw_email = msg_data[0][1]
        msg = email.message_from_bytes(raw_email, policy=policy.default)
        self.logger.info(f"Contenu de l'email {email_id} correctement lu.")
        message = self._decode_email_body(msg)
        if not get_attchment:
            return {"message": message}
        attachments = self.__get_attachment(msg)
        return {"message": message, "attachment": attachments}

    def logout(self) -> None:
        """Ferme la connexion avec le serveur mail."""
        self.__connection.logout()



"""
Module providing database DDL (Data Definition Language) operations.

This module defines abstract and concrete classes for managing DDL operations
such as creating, altering, truncating, and dropping tables and views,
as well as creating indexes for both Microsoft SQL Server and PostgreSQL databases.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from sqlalchemy import Column, Engine, Index, Selectable, Table, exc, schema, text

if TYPE_CHECKING:
    from .logger import Logger


class DdlOperations(ABC):
    """Abstract base class for DDL operations."""

    def __init__(self, engine: Engine, logger: Logger) -> None:
        """
        Initialize the DDL operations handler.

        :param engine: SQLAlchemy engine connected to the target database.
        :param logger: Logger instance.
        """
        self.engine = engine
        self.logger = logger

    def quote(self, name: str) -> str:
        """Quote database object names."""
        return self.engine.dialect.identifier_preparer.quote(name)

    def execute(self, query: str) -> None:
        """Execute a raw SQL query."""
        try:
            with self.engine.begin() as conn:
                conn.execute(text(query))
                self.logger.debug(f"Executed query: {query}")
        except exc.SQLAlchemyError as e:
            self.logger.debug(f"Error executing query: {query} : {e}")
            raise

    def create_table(self, table: Table) -> None:
        """Create a table."""
        if self.engine.dialect.has_table(self.engine.connect(), table.name, schema=table.schema):
            table.metadata.reflect(bind=self.engine, extend_existing=True, schema=table.schema, only=[table.name])
            self.logger.info(f"Table {table.schema}.{table.name} already exists.")
            return
        sql_statement = str(schema.CreateTable(table).compile(self.engine))
        try:
            self.execute(sql_statement)
            table.metadata.reflect(bind=self.engine, extend_existing=True, schema=table.schema, only=[table.name])
            self.logger.info(f"Created table {table.schema}.{table.name}.")
        except exc.SQLAlchemyError as e:
            self.logger.error(f"Error creating table {table.schema}.{table.name}: {e}.")
            raise

    def drop_table(self, table: Table) -> None:
        """Drop a table."""
        if not self.engine.dialect.has_table(self.engine.connect(), table.name, schema=table.schema):
            self.logger.info(f"Table {table.schema}.{table.name} does not exist.")
            return
        sql_statement = str(schema.DropTable(table).compile(self.engine))
        try:
            self.execute(sql_statement)
            self.logger.info(f"Dropped table {table.schema}.{table.name}.")
        except exc.SQLAlchemyError as e:
            self.logger.error(f"Error dropping table {table.schema}.{table.name}: {e}.")
            raise

    def truncate_table(self, table: Table) -> None:
        """Truncate a table."""
        if not self.engine.dialect.has_table(self.engine.connect(), table.name, schema=table.schema):
            self.logger.info(f"Table {table.schema}.{table.name} does not exist.")
            return
        sql_statement = f"TRUNCATE TABLE {self.quote(table.schema)}.{self.quote(table.name)}"
        try:
            self.execute(sql_statement)
            self.logger.info(f"Truncated table {table.schema}.{table.name}.")
        except exc.SQLAlchemyError as e:
            self.logger.error(f"Error truncating table {table.schema}.{table.name}: {e}.")
            raise

    @abstractmethod
    def _add_column_sql(self, table: Table, column: Column) -> str:
        """Generate SQL for adding a column."""

    def add_column(self, table: Table, column: Column) -> None:
        """Add a column to a table."""
        if column.name in table.c:
            self.logger.info(f"Column {column.name} already exists in table {table.schema}.{table.name}.")
            return
        sql_statement = self._add_column_sql(table, column)
        try:
            self.execute(sql_statement)
            table.append_column(column)
            self.logger.info(f"Added column {column.name} to table {table.schema}.{table.name}.")
        except exc.SQLAlchemyError as e:
            self.logger.error(f"Error adding column {column.name} to table {table.schema}.{table.name}: {e}.")
            raise

    @abstractmethod
    def _create_view_sql(self, view: Table, definition: Selectable, *, or_replace: bool=True) -> str:
        """Generate SQL for creating a view."""

    def create_view(self, view: Table, definition: Selectable, *, or_replace: bool=True) -> None:
        """Create a view."""
        if self.engine.dialect.has_table(self.engine.connect(), view.name, schema=view.schema) and not or_replace:
            view.metadata.reflect(bind=self.engine, views=True, extend_existing=True, schema=view.schema,
                                  only=[view.name])
            self.logger.info(f"View {view.schema}.{view.name} already exists")
            return
        sql_statement = self._create_view_sql(view, definition, or_replace)
        try:
            self.execute(sql_statement)
            view.metadata.reflect(bind=self.engine, views=True, extend_existing=True, schema=view.schema,
                                  only=[view.name])
            self.logger.info(f"Created view {view.schema}.{view.name}.")
        except exc.SQLAlchemyError as e:
            self.logger.error(f"Error creating view {view.schema}.{view.name}: {e}.")
            raise

    def create_index(self, table: Table, columns_name: list, index_name: str, *, unique: bool=False) -> None:
        """Create an index on multiple table columns."""
        if index_name in [idx.name for idx in table.indexes]:
            self.logger.info(f"Index {index_name} already exists on table {table.schema}.{table.name}.")
            return
        for column_name in columns_name:
            if column_name not in table.c:
                msg = f"Column {column_name} does not exist in table {table.schema}.{table.name}."
                raise ValueError(msg)
        index = Index(index_name, *[table.c[column_name] for column_name in columns_name], unique=unique)
        try:
            index.create(self.engine)
            self.logger.info(
                f"Created index {index_name} on columns {', '.join(columns_name)} of table {table.schema}.{table.name}.")
        except exc.SQLAlchemyError as e:
            self.logger.error(
                f"Error creating index {index_name} on columns {', '.join(columns_name)} of table {table.schema}.{table.name}: {e}.")
            raise

class MssqlDdlOperations(DdlOperations):
    """Concrete class for MSSQL DDL operations."""

    def _add_column_sql(self, table: Table, column: Column) -> str:
        column_definition = f"{self.quote(column.name)} {column.type.compile(self.engine.dialect)}"
        return f"ALTER TABLE {self.quote(table.schema)}.{self.quote(table.name)} ADD {column_definition}"

    def _create_view_sql(self, view: Table, definition: Selectable, *, or_replace: bool=True) -> str:
        if or_replace:
            sql_statement = f"CREATE OR ALTER VIEW {self.quote(view.schema)}.{self.quote(view.name)} AS {definition.compile(compile_kwargs={'literal_binds': True})}"
        else:
            sql_statement = f"CREATE VIEW {self.quote(view.schema)}.{self.quote(view.name)} AS {definition.compile(compile_kwargs={'literal_binds': True})}"
        return sql_statement

class PostgresqlDdlOperations(DdlOperations):
    """Concrete class for PostgreSQL DDL operations."""

    def _add_column_sql(self, table: Table, column: Column) -> str:
        column_definition = f"{self.quote(column.name)} {column.type.compile(self.engine.dialect)}"
        return f"ALTER TABLE {self.quote(table.schema)}.{self.quote(table.name)} ADD COLUMN {column_definition}"

    def _create_view_sql(self, view: Table, definition: Selectable, *, or_replace: bool=True) -> str:
        if or_replace:
            sql_statement = f"CREATE OR REPLACE VIEW {self.quote(view.schema)}.{self.quote(view.name)} AS {definition.compile(compile_kwargs={'literal_binds': True})}"
        else:
            sql_statement = f"CREATE VIEW {self.quote(view.schema)}.{self.quote(view.name)} AS {definition.compile(compile_kwargs={'literal_binds': True})}"
        return sql_statement

def create_ddl_operations(engine: Engine, logger: Logger) -> MssqlDdlOperations | PostgresqlDdlOperations:
    """Create the appropriate DDL operations class."""
    dialect = engine.dialect.name
    if dialect == "mssql":
        return MssqlDdlOperations(engine, logger)
    if dialect == "postgresql":
        return PostgresqlDdlOperations(engine, logger)
    msg = f"DDL operations for database type '{dialect}' are not implemented."
    raise NotImplementedError(msg)



"""
Module `git_handler` pour gérer les dépôts Git et les environnements virtuels.

Ce module fournit la classe `GitHandler` permettant de :
- Cloner ou mettre à jour des dépôts GitHub à l'aide d'un token d'accès.
- Vérifier les permissions d'accès à un chemin donné.
- Créer et configurer un environnement virtuel Python dans un dépôt, avec installation
  des dépendances et droits d'exécution sur un script de lancement.
"""
from __future__ import annotations

import os
import subprocess
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from git import Repo
from github import Github

if TYPE_CHECKING:
    from .logger import Logger
from .timer import timer


class GitHandler:
    """
    Classe pour gérer les dépôts GitHub et les environnements virtuels Python.

    Attributs:
        token (str): Token d'accès GitHub utilisé pour l'authentification.
        github (Github): Instance de la bibliothèque PyGithub pour les interactions avec GitHub.
        logger (Logger): Objet Logger pour la journalisation.
    """

    def __init__(self, token: str, logger: Logger) -> None:
        """
        Initialise le gestionnaire GitHandler avec un token GitHub et un logger.

        :param token: Token d'accès GitHub.
        :param logger: Objet Logger pour journaliser les actions et les erreurs.
        """
        self.token = token
        self.github = Github(token)
        self.logger = logger

    def check_permissions(self, path: str) -> None:
        """Vérifie que le script dispose des permissions nécessaires sur le chemin spécifié."""
        if not os.access(path, os.R_OK):
            msg = f"Le script n'a pas la permission de lecture sur le chemin: {path}"
            raise PermissionError(msg)

    @timer
    def clone_or_update(self, destination_path: str, repository: str, branch: str | None = None) -> None:
        """Clone le repo s'il existe pas localement, sinon pull la dernière version."""
        self.logger.info(f"Connexion au repo {repository}")
        url = f"https://{self.token}@github.com/{repository}.git"
        try:
            path = Path(destination_path)
            if path.exists():
                self.check_permissions(destination_path)
                repo = Repo(destination_path)

                self.logger.debug("Mise à jour du remote origin avec le token.")
                repo.remotes.origin.set_url(url)

                repo.remotes.origin.fetch()
                if branch:
                    self.logger.debug(f"Changement de branche : {branch}")
                    repo.git.checkout(branch)
                repo.remotes.origin.pull()
                self.logger.info(f"Le repo {repository} a été mis à jour.")
            else:
                self.logger.info(f"Clonage du repo {repository} dans {destination_path}...")
                if branch:
                    Repo.clone_from(url, destination_path, branch=branch)
                else:
                    Repo.clone_from(url, destination_path)
                self.logger.info(f"Le repo {repository} a été cloné avec succès.")
        except Exception as e:
            self.logger.error(f"Erreur lors du clonage du repo {repository}: {e}")
            raise

    @timer
    def setup_virtualenv(self, repo_path: str) -> None:
        """Crée et installe d'un environnement virtuel dans le repo."""
        repo = Path(repo_path)
        venv_path = repo / "venv"
        requirements_path = repo / "requirements.txt"
        launch_path = repo / "launch.sh"

        self.logger.info(f"Création de l'environnement virtuel dans {venv_path}...")
        try:
            subprocess.run([sys.executable, "-m", "venv", str(venv_path)], check=True) # noqa: S603
        except Exception as e:
            self.logger.error(f"Erreur lors de la création du venv: {e}")
            raise
        if requirements_path.exists():
            self.logger.info("Installation des dépendances à partir de requirements.txt...")
            pip_executable = venv_path / (Path("Scripts") / "pip.exe" if os.name == "nt" else Path("bin") / "pip")
            subprocess.run([str(pip_executable), "install", "-r", str(requirements_path)], check=True) # noqa: S603
        else:
            self.logger.warning("Aucun fichier requirements.txt trouvé, installation ignorée.")

        # Attribution des droits d exécution au script launch.sh (si présent)
        if launch_path.exists():
            self.logger.info("Attribution des droits d'exécution sur launch.sh...")
            try:
                launch_path.chmod(0o755)
                self.logger.info("Droits d'exécution appliqués avec succès sur launch.sh.")
            except PermissionError as e:
                self.logger.error(f"Impossible de modifier les permissions de launch.sh : {e}")
        else:
            self.logger.warning("Aucun fichier launch.sh trouvé dans le dépôt.")
        self.logger.info("Environnement virtuel configuré avec succès.")



"""
Package adsToolBox.

Contient les modules pour :
- Gestion des bases de données et pipeline
- Gestion des fichiers
- Logging
- Outils Odoo
- Gestion Repo Git
- Journalisation et Chronomètre
- Gestion de l'environnement
- Objet de lecture de mail
"""

from .cdc import ChangeDataCapture
from .data_comparator import DataComparator
from .data_factory import DataFactory
from .db_mssql import DbMssql
from .db_mysql import DbMysql
from .db_pgsql import DbPgsql
from .file_handler import FileHandler
from .git_handler import GitHandler
from .global_config import get_public_ip, set_timer
from .load_env import Env
from .logger import Logger
from .mail_reader import MailReader
from .odoo import OdooConnector
from .pipeline import Pipeline
from .timer import get_timer, now, set_timezone, timer

__all__ = [
    "ChangeDataCapture",
    "DataComparator",
    "DataFactory",
    "DbMssql",
    "DbMysql",
    "DbPgsql",
    "Env",
    "FileHandler",
    "GitHandler",
    "Logger",
    "MailReader",
    "OdooConnector",
    "Pipeline",
    "get_public_ip",
    "get_timer",
    "now",
    "set_timer",
    "set_timezone",
    "timer",
]



"""
Module de logger.

Contient la classe Logger pour la gestion des logs avec :
- affichage coloré en console,
- écriture dans un fichier rotatif,
- insertion optionnelle dans une base de données.

Inclut également CustomFormatter pour la coloration des logs en console.
"""
from __future__ import annotations

import logging
import sys
import traceback
from datetime import datetime, timezone
from logging import LogRecord
from logging.handlers import TimedRotatingFileHandler
from pathlib import Path
from typing import TYPE_CHECKING
from uuid import uuid4

from .db_mssql import DbMssql
from .db_mysql import DbMysql
from .db_pgsql import DbPgsql

if TYPE_CHECKING:
    from .data_factory import DataFactory

class CustomFormatter(logging.Formatter):
    """
    Un formateur personnalisé pour les logs permettant l'affichage en différentes couleurs.

    - DEBUG : vert
    - INFO : gris
    - WARNING : jaune
    - ERROR : rouge
    Seuls les messages print dans la console ont leurs couleurs changées
    """

    grey = "\x1b[37m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    light_green = "\x1b[92m"
    blue = "\x1b[94m"
    reset = "\x1b[0m"

    def __init__(self, base_format: str) -> None:
        """
        Initialise le CustomFormatter avec un format spécifique.

        :param base_format: Format de base pour les logs
        """
        super().__init__()
        self.base_format = base_format
        self.FORMATS = {
            logging.DEBUG: self.light_green + self.base_format + self.reset,
            logging.INFO: self.blue + self.base_format + self.reset,
            logging.WARNING: self.yellow + self.base_format + self.reset,
            logging.ERROR: self.red + self.base_format + self.reset,
            logging.CRITICAL: self.bold_red + self.base_format + self.reset,
            None: base_format,
        }

    def format(self, record: LogRecord) -> str:
        """Renvoie le message formaté avec les couleurs en fonction du niveau de log."""
        log_fmt = self.FORMATS.get(record.levelno, self.FORMATS[None])
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class Logger:
    """
     Classe `Logger` pour gérer la journalisation dans les fichiers, la console et une base de données PostgreSQL ou MSSQL.

    Elle permet d'écrire les logs dans un fichier rotatif, dans la console avec des couleurs personnalisées,
    et également de les stocker dans une base de données, si spécifié.
    """

    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL

    def __init__(
            self,
            log_level: int=INFO,
            logger_name: str= "AdsLogger",
            table_log_name: str= "LOGS",
            table_log_details_name: str="LOGS_details",
            *, timestamp_display: bool=True, name_display: bool=True,
    ) -> None:
        """
        Logger gère la journalisation dans les fichiers, la console et en base de données.

        Écrit dans la console avec des couleurs différentes, en fichier avec un backup de 10 jours et en base avec une
        connexion spécifiée
        :param log_level: le niveau de log courant, un message de debug ne sera pas affiché si le logger a un niveau info
        :param logger_name: le nom du logger
        :param table_log_name: le nom de la table de log
        :param table_log_details_name: le nom de la seconde table de log
        :param timestamp_display: booléen qui indique si l'horodatage doit être affiché
        :param name_display: booléen qui indique si le nom du logger doit être affiché
        """
        self.log_level_global = log_level
        self.log_level_base = log_level
        self.logger = self.__setup_logger(logger_name, timestamp_display=timestamp_display, name_display=name_display)
        self.connection = None
        self.db_logging_enabled = False
        self.table_log_name = table_log_name
        self.table_log_details_name = table_log_details_name
        self.creation_date = datetime.now(timezone.utc)
        self.job_key = uuid4()

    def __setup_logger(self, logger_name: str, *, timestamp_display: bool, name_display: bool) -> logging.Logger:
        """
        Configure le logger avec un gestionnaire pour les fichiers et un pour la console.

        :param logger_name: le nom du logger à configurer
        :return: le logger configuré
        """
        logger = logging.getLogger(logger_name)
        logger.setLevel(self.log_level_global)

        # Construire le format en fonction des paramètres
        base_format = "%(levelname)s - %(message)s"
        if name_display:
            base_format = "%(name)s - " + base_format
        if timestamp_display:
            base_format = "%(asctime)s - " + base_format

        file_handler = TimedRotatingFileHandler(filename="log.txt", when="D", interval=1, backupCount=10)
        file_handler.setLevel(self.log_level_global)
        file_handler.setFormatter(logging.Formatter(base_format))

        console_handler = logging.StreamHandler()
        console_handler.setLevel(self.log_level_global)
        console_handler.setFormatter(CustomFormatter(base_format))

        logger.addHandler(file_handler)
        logger.addHandler(console_handler)

        return logger

    def set_connection(self, ads_connection: DataFactory, log_level: int) -> None:
        """
        Attribue la connexion passée en paramètre au logger.

        :param log_level: Le niveau de log pour l'insertion en base
        :param ads_connection: La connexion ads en question
        """
        self.connection = ads_connection
        self.db_logging_enabled = True
        self.log_level_base = log_level

    def _get_connection(self): # noqa: ANN202
        """
        Récupère la connexion associée au logger si elle existe.

        :return: la connexion
        """
        if isinstance(self.connection, (DbMssql, DbMysql, DbPgsql)):
            return self.connection.connection
        return self.connection

    def create_logs_tables(self) -> None:
        """
        Crée les tables de logs dans la base de données.

        Les tables créées sont :
        - `table_log_name` : Contient les logs principaux.
        - `table_log_details_name` : Contient les détails des logs.
        """
        if not self.connection or not self.db_logging_enabled:
            return
        if isinstance(self.connection, DbPgsql):
            logs_query = f"""
            CREATE TABLE {self.table_log_name} (
                start_time TIMESTAMP,
                end_time TIMESTAMP,
                job_key UUID PRIMARY KEY,
                full_path VARCHAR(4000),
                status VARCHAR(50),
                message VARCHAR(4000)
            );"""
            details_query = f"""
            CREATE TABLE {self.table_log_details_name} (
                date TIMESTAMP,
                job_key UUID,
                log_level VARCHAR(20),
                message VARCHAR(4000),
                PRIMARY KEY (job_key, date)
            );"""
        elif isinstance(self.connection, DbMssql):
            logs_query = f"""
            CREATE TABLE {self.table_log_name} (
                start_time DATETIME,
                end_time DATETIME,
                job_key UNIQUEIDENTIFIER PRIMARY KEY,
                full_path NVARCHAR(4000),
                status NVARCHAR(50),
                message NVARCHAR(4000)
            );"""
            details_query = f"""
            CREATE TABLE {self.table_log_details_name} (
                date DATETIME,
                job_key UNIQUEIDENTIFIER,
                log_level NVARCHAR(20),
                message NVARCHAR(4000),
                PRIMARY KEY (job_key, date)
            );"""
        elif isinstance(self.connection, DbMysql):
            logs_query = f"""
            CREATE TABLE {self.table_log_name} (
                start_time DATETIME,
                end_time DATETIME,
                job_key CHAR(36) PRIMARY KEY,
                full_path VARCHAR(4000),
                status VARCHAR(50),
                message VARCHAR(4000)
            );"""
            details_query = f"""
            CREATE TABLE {self.table_log_details_name} (
                date DATETIME,
                job_key CHAR(36),
                log_level VARCHAR(20),
                message VARCHAR(4000),
                PRIMARY KEY (job_key, date)
            );"""
        else:
            msg = "Connexion à une base de données non supportée pour la création de tables."
            self.error(msg)
            raise NotImplementedError(msg)
        try:
            with self._get_connection().cursor() as cursor:
                cursor.execute(logs_query)
                cursor.execute(details_query)
                self._get_connection().commit()
                self.info(f"Table {self.table_log_name} et {self.table_log_details_name} créées.", insert=False)
        except Exception as e: # noqa: BLE001
            self._get_connection().rollback()
            self.error(f"Erreur lors de la création des tables de logs: {e}")

    def log_close(self, status: str, message: str) -> None:
        """
        Lance l'insertion dans la table de log principale et désactive les logs.

        :param status: Le statut du log à insérer
        :param message: Le message à insérer
        """
        self.info("Fermeture des logs en cours...", insert=False)
        self._insert_logs_table(status, message)
        self.disable()

    def _insert_logs_table(self, status: str, message: str) -> None:
        """
        Insère les données dans la table de logs principales.

        :param status: Statut de l'évènement
        :param message: Message de log
        """
        if not self.connection or not self.db_logging_enabled:
            self.warning("Pas de connexion valide ou logs désactivés.")
            return
        full_path = Path(sys.argv[0]).resolve().relative_to(Path.cwd().parent)
        try:
            with self._get_connection().cursor() as cursor:
                cursor.execute(
                    f"""
                    INSERT INTO {self.table_log_name} (job_key, start_time, end_time, full_path, status, message)
                    VALUES (%s, %s, %s, %s, %s, %s)""",
                    (
                        str(self.job_key),
                        self.creation_date,
                        datetime.now(timezone.utc),
                        full_path,
                        status,
                        message[:4_000],
                    ),
                )
                self._get_connection().commit()
                self.info(f"Logs insérés dans la table de {self.table_log_name}.", insert=False)
        except Exception as e:
            self._get_connection().rollback()
            self.error(f"Échec de l'insertion dans la table de logs {self.table_log_name}: {e}",
                       insert=False)
            raise


    def _insert_logs_table_details(self, message: str, log_level: int) -> None:
        """
        Insère les données dans la table de logs détails.

        :param message: Message de log
        :param log_level: Le niveau de log
        """
        if not self.connection or not self.db_logging_enabled:
            return
        try:

            with self._get_connection().cursor() as cursor:

                cursor.execute(
                    f"""
                    INSERT INTO {self.table_log_details_name} (job_key, date, log_level, message)
                    VALUES (%s, %s, %s, %s)
                    """,
                    (
                        str(self.job_key),
                        datetime.now(timezone.utc),
                        log_level,
                        message[:4_000],
                    ),
                )
                self._get_connection().commit()
        except Exception as e:
            self._get_connection().rollback()
            self.error(f"Échec de l'insertion dans la table de détails {self.table_log_details_name}: {e}",
                       insert=False)
            raise

    def info(self, message: str, *, insert: bool=True) -> None:
        """
        Log un message avec le niveau info.

        :param insert: True si le message est à insérer en base
        :param message: le message à logguer
        """
        if self.log_level_global <= self.INFO:
            self.logger.info(message)
        if self.log_level_base <= self.INFO and insert:
            self._insert_logs_table_details(message, self.INFO)

    def error(self, message: str, *, insert: bool=True) -> None:
        """
        Log un message avec le niveau erreur.

        :param insert: True si le message est à insérer en base
        :param message: le message à logguer
        """
        if self.log_level_global <= self.ERROR:
            self.logger.error("*"*50)
            self.logger.error(message.replace("\n", " "))
            self.logger.error("*"*50)
        if self.log_level_base <= self.ERROR and insert:
            trace = traceback.format_exc()
            full_msg = f"{message}\nTraceback: {trace}" if "NoneType: None" not in trace else message
            self._insert_logs_table_details(full_msg, self.ERROR)

    def debug(self, message: str) -> None:
        """
        Log un message avec le niveau debug.

        :param message: le message à logguer
        """
        if self.log_level_global <= self.DEBUG:
            self.logger.debug(message)
        if self.log_level_base <= self.DEBUG:
            self._insert_logs_table_details(message, self.DEBUG)

    def warning(self, message: str) -> None:
        """
        Log un message avec le niveau warning.

        :param message: le message à logguer
        """
        if self.log_level_global <= self.WARNING:
            self.logger.warning(message)
        if self.log_level_base <= self.WARNING:
            self._insert_logs_table_details(message, self.WARNING)

    def custom_log(self, log_level: int, message: str) -> None:
        """
        Log un message avec un niveau custom.

        :param log_level: le niveau de log
        :param message: le message
        """
        if log_level >= self.log_level_global:
            self.logger.log(log_level, message)
        if log_level >= self.log_level_base:
            self._insert_logs_table_details(message, log_level)

    def enable(self, level_logger: int=INFO, level_for_base: int=INFO, *, generate_new_key: bool=False) -> None:
        """
        Active la journalisation et définit le niveau de log.

        :param generate_new_key: True si on doit définir un nouveau job_key
        :param level_for_base: Le niveau de log pour l'insertion en base
        :param level_logger: Le niveau de log pour l'affichage en console/fichier
        """
        self.logger.setLevel(level_logger)
        self.log_level_global = level_logger
        self.log_level_base = level_for_base
        self.db_logging_enabled = True
        if generate_new_key:
            self.job_key = uuid4()

    def disable(self) -> tuple[int, int]:
        """
        Désactive la journalisation en réglant le niveau de log au niveau CRITICAL et en désactivant l'insertion en base.

        :return : Les anciens niveaux de logs avant la désactivation
        """
        old_global = self.log_level_global
        old_base = self.log_level_base
        self.logger.setLevel(self.CRITICAL)
        self.log_level_global = self.CRITICAL
        self.log_level_base = self.CRITICAL
        self.db_logging_enabled = False
        return old_global, old_base



"""
Module file_handler.

Ce module fournit la classe `FileHandler` pour gérer les opérations sur les fichiers
en local et via des partages SMB ou FTP. Il inclut la lecture et l'écriture en batchs,
la vérification de l'existence de fichiers et de dossiers, la suppression et la création
de fichiers vides, ainsi que le calcul de la taille des fichiers et des répertoires.
"""
from __future__ import annotations

import functools
import hashlib
import time
import timeit
from datetime import datetime, timezone
from pathlib import Path
from typing import TYPE_CHECKING, Callable, ParamSpec, TypeVar
from uuid import uuid4

import fsspec

from adsToolBox import DbMssql, DbMysql, DbPgsql

from .timer import get_timer, timer

if TYPE_CHECKING:
    from collections.abc import Generator, Iterable

    from .data_factory import DataFactory
    from .logger import Logger

OCTETS_PAR_KO = 1024.0
ASCII_MAX = 127

P = ParamSpec("P")
R = TypeVar("R")

def retry_on_failure(func: Callable) -> Callable:
    """
    Décorateur qui réessaie une méthode plusieurs fois en cas d'erreur.

    La classe appelante doit définir :
        - retry_count (int)
        - retry_delay (float)
        - logger (Logger ads)
    """
    @functools.wraps(func)
    def wrapper(self: FileHandler, *args: P.args, **kwargs: P.kwargs) -> R:
        for attempt in range(1, self.retry_count + 1):
            try:
                return func(self, *args, **kwargs)
            except Exception as e: # noqa: PERF203
                self.logger.warning(
                    f"Tentative {attempt}/{self.retry_count} échouée : {e}",
                )
                if attempt >= self.retry_count:
                    self.logger.error(
                        f"Échec définitif après {self.retry_count} tentatives ({int(self.retry_count * self.retry_delay)}s).",
                    )
                    raise
                time.sleep(self.retry_delay)
        msg = "Erreur interne de retry_on_failure, attention retry_count à 0 siginie 0 tentatives."
        raise RuntimeError(msg)
    return wrapper

class FileHandler:
    """
    Gestionnaire de fichiers local et distant.

    Cette classe permet de manipuler des fichiers et des répertoires en supportant les opérations suivantes :

    - Lecture de fichiers en batchs (`read_file`)
    - Écriture de fichiers en batchs (`write_file`)
    - Transfert de fichier entre différents systèmes ('local', 'SMB', 'Azure', 'FTP', etc)
    - Vérification après transfert (tailles ou checksum via MD5)
    - Vérification de l'existence de fichiers et dossiers (`file_exists`, `directory_exists`)
    - Suppression et création de fichiers (`remove_file`, `create_empty_file`)
    - Nettoyage de texte (`clean_text`)
    - Calcul de la taille d'un répertoire (`_get_dir_size`)
    - Collecte et traitement des fichiers selon un filtre (`_collect_files`, `_process_files`, `disk_check`)
    - Attente de la présence d'un fichier avec retry (`wait_for_file`)

    Attributs :
        logger (Logger): Système de log injecté.
        fs (fsspec.AbstractFileSystem): Instance de système de fichiers (local, smb, sftp, etc.).
        batch_size (int): Taille maximale des chunks pour lecture/écriture.
        retry_count (int): Nombre de tentatives avant échec.
        retry_delay (float): Délai (en secondes) entre les tentatives.
    """

    def __init__(
            self,
            logger: Logger,
            fs_url: str | None = None,
            fs_kwargs: dict | None = None,
            batch_size: int = 4096,
            retry_count: int = 3,
            retry_delay: float = 2.0,
    )-> None:
        """
        Initialise le gestionnaire de fichiers universel basé sur fsspec.

        :param logger: Instance de Logger.
        :param fs_url: URL fsspec (ex: 'file:///', 'smb://server/share', 'sftp://user:pass@host/').
                       Si None, le backend local est utilisé.
        :param fs_kwargs: Paramètres optionnels pour fsspec (ex: credentials, host, port...).
        :param batch_size: Taille des chunks de lecture/écriture.
        :param retry_count: Nombre de tentatives avant échec.
        :param retry_delay: Délai entre tentatives (en secondes).
        """
        self.logger = logger
        self.batch_size = batch_size
        self.retry_count = retry_count
        self.retry_delay = retry_delay
        self._set_connection(fs_url or "file:///", fs_kwargs or {})

    @retry_on_failure
    def _set_connection(self, fs_url: str, fs_kwargs: dict) -> None:
        self.fs, self.root_url = fsspec.core.url_to_fs(fs_url, **fs_kwargs)
        self.logger.debug(f"FileHandler initialisé avec backend '{self.fs.protocol}' (url={fs_url})")

    def set_retry_params(self, retry_count: int, retry_delay: float) -> None:
        """Change les paramètres de nombre de retry et de délai."""
        self.retry_count = retry_count
        self.retry_delay = retry_delay

    @retry_on_failure
    def read_file(self, file_path: str, mode: str="rb", encoding: str | None = None) -> Generator[str | bytes, None, None]:
        """
        Lit un fichier distant ou local via fsspec, en streaming par blocs.

        Supporte ces modes : 'r', 'rb'.
        """
        timer_start = timeit.default_timer()
        try:
            with self.fs.open(file_path, mode=mode, encoding=encoding) as f:
                while True:
                    chunk = f.read(self.batch_size)
                    if not chunk:
                        break
                    yield chunk
            self.logger.info(f"Succès: lecture du fichier '{file_path}' ({self.fs.protocol})")
        except Exception as e:
            self.logger.error(f"Erreur lors de la lecture du fichier '{file_path}' via {self.fs.protocol}: {e}")
            raise
        finally:
            if get_timer():
                elapsed_time = timeit.default_timer() - timer_start
                self.logger.info(f"Temps d'exécution de read_file: {elapsed_time:.4f} secondes.")

    @retry_on_failure
    @timer
    def write_file(self, file_path: str, content: Iterable[str | bytes], mode: str="wb") -> None:
        """
        Écrit du contenu dans un fichier local ou distant via fsspec.

        Supporte ces modes : 'w', 'wb', 'a', 'ab', 'x', 'xb'.
        """
        encoding = "utf-8" if "b" not in mode else None
        is_append = "a" in mode
        is_exclusvie = "x" in mode
        sep = "/" if "/" in file_path else "\\"
        dir_part, _, _ = file_path.rpartition(sep)
        if is_exclusvie and self.fs.exists(file_path):
            msg = f"Le fichier '{file_path}' existe déjà (mode 'x')."
            self.logger.error(msg)
            raise FileExistsError(msg)
        if is_append:
            target_path = file_path
        else:
            temp_filename = f"tmp_{uuid4().hex}"
            target_path = f"{dir_part}{sep}{temp_filename}" if dir_part else temp_filename
        try:
            with self.fs.open(target_path, mode=mode, encoding=encoding) as f:
                for chunk in content:
                    data = chunk.encode(encoding or "utf-8") if isinstance(chunk, str) and "b" in mode else chunk
                    for i in range(0, len(data), self.batch_size):
                        f.write(data[i:i + self.batch_size])
            if not is_append:
                self.fs.mv(target_path, file_path)
            self.logger.info(f"Succès: contenu écrit dans '{file_path}' ({self.fs.protocol})")
        except Exception as e:
            if not is_append and self.fs.exists(target_path):
                try:
                    self.fs.rm(target_path)
                    self.logger.warning(f"Fichier temporaire supprimé après erreur: {target_path}.")
                except Exception as cleanup_error: # noqa: BLE001
                    self.logger.warning(f"Impossible de supprimer le fichier temporaire: {target_path}: {cleanup_error}.")
            self.logger.error(f"Erreur lors de l'écriture du fichier '{file_path}': {e}")
            raise

    def _read_with_checksum(self, path: str, mode: str="rb") -> tuple[Generator[bytes, None, None], Callable[[], str]]:
        """Lit un fichier en streaming et calcule le checksum en même temps."""
        h = hashlib.new("md5") # noqa: S324
        def gen() -> Generator[bytes, None, None]:
            with self.fs.open(path, mode=mode) as f:
                while True:
                    chunk = f.read(self.batch_size)
                    if not chunk:
                        break
                    h.update(chunk)
                    yield chunk
        def get_checksum() -> str:
            return h.hexdigest()
        return gen(), get_checksum

    def write_with_checksum(self, path: str, content: Iterable[bytes], mode: str= "wb") -> str:
        """Écrit un fichier en streaming et calcule le checksum en même temps."""
        h = hashlib.new("md5") # noqa: S324
        is_append = "a" in mode
        is_exclusive = "x" in mode
        sep = "/" if "/" in path else "\\"
        dir_part, _, _ = path.rpartition(sep)
        if is_exclusive and self.fs.exists(path):
            msg = f"Le fichier '{path}' existe déjà (mode 'x')."
            self.logger.error(msg)
            raise FileExistsError(msg)
        if is_append:
            target_path = path
        else:
            temp_filename = f"tmp_{uuid4().hex}"
            target_path = f"{dir_part}{sep}{temp_filename}" if dir_part else temp_filename
        try:
            with self.fs.open(target_path, mode=mode) as f:
                for chunk in content:
                    h.update(chunk)
                    for i in range(0, len(chunk), self.batch_size):
                        f.write(chunk[i:i + self.batch_size])
            if not is_append:
                self.fs.mv(target_path, path)
        except Exception as e:
            if not is_append and self.fs.exists(target_path):
                try:
                    self.fs.rm(target_path)
                    self.logger.warning(f"Fichier temporaire supprimé après erreur: {target_path}.")
                except Exception as cleanup_error: # noqa: BLE001
                    self.logger.warning(f"Impossible de supprimer le fichier temporaire: {target_path}: {cleanup_error}.")
            self.logger.error(f"Erreur lors de l'écriture du fichier '{path}': {e}")
            raise
        return h.hexdigest()

    @timer
    @retry_on_failure
    def transfer_file(
            self,
            src_path: str,
            dst_path: str,
            dst_file_handler: FileHandler | None = None, # dst_file_handler doit se trouver ici si on veut garder sa valeur None par défaut
            mode: str="w",
            *, fastcheck: bool = True,
    ) -> bool:
        """
        Copie un fichier puis vérifie l'intégrité.

        Les modes disponibles sont 'w' qui écrase la destination, 'a' qui y ajoute le contenu et 'x' qui lève une erreur
        si la destination existe déjà.
        """ # noqa: D401
        dst_file_handler = dst_file_handler or self
        if "b" not in mode:
            mode += "b"
        self.logger.debug(f"Transfert de {src_path} vers {dst_path} (fastcheck={fastcheck}).")
        log_level, base_level = self.logger.log_level_global, self.logger.log_level_base
        try:
            if fastcheck:
                self.logger.disable()
                dst_file_handler.write_file(dst_path, self.read_file(src_path), mode)
                self.logger.enable(log_level, base_level)
                src_size = self.fs.info(src_path)["size"]
                dst_size = dst_file_handler.fs.info(dst_path)["size"]
                if src_size != dst_size:
                    self.logger.error(f"Transfert KO (fastcheck) - tailles différentes: {src_size} != {dst_size}")
                    return False
                self.logger.info(f"Transfert OK (fastcheck) - tailles identiques: {src_size} != {dst_size}")
                return True
            content, get_checksum_src = self._read_with_checksum(src_path)
            checksum_dst = dst_file_handler.write_with_checksum(dst_path, content)
            checksum_src = get_checksum_src()
            if checksum_src != checksum_dst:
                self.logger.error(f"Transfert KO - checksum différent: {checksum_src} != {checksum_dst}")
                return False
            self.logger.info(f"Transfert OK - checksum identique: {checksum_src} != {checksum_dst}")
        except Exception as e:
            self.logger.enable(log_level, base_level)
            self.logger.error(f"Erreur lors du transfert {src_path} -> {dst_path}: {e}")
            raise
        return True

    @retry_on_failure
    @timer
    def list_dir(self, dir_path: str) -> list[str]:
        """Liste le contenu d'un dossier."""  # noqa: D401
        try:
            entries = self.fs.ls(dir_path, detail=False)
            return [Path(e).name for e in entries]
        except Exception as e:
            self.logger.error(f"Erreur lors du listage de '{dir_path}': {e}")
            raise

    @retry_on_failure
    @timer
    def file_exists(self, file_path: str) -> bool:
        """Vérifie si un fichier existe."""
        try:
            return self.fs.exists(file_path) and self.fs.isfile(file_path)
        except Exception as e:
            self.logger.error(f"Erreur lors de la vérification du fichier '{file_path}': {e}")
            raise

    @retry_on_failure
    @timer
    def directory_exists(self, dir_path: str) -> bool:
        """Vérifie si un répertoire existe."""
        try:
            return self.fs.exists(dir_path) and self.fs.isdir(dir_path)
        except Exception as e:
            self.logger.error(f"Erreur lors de la vérification du dossier '{dir_path}': {e}")
            raise

    @retry_on_failure
    @timer
    def remove_file(self, file_path: str) -> bool:
        """Supprime un fichier."""
        log_level, base_level = self.logger.disable()
        try:
            if self.file_exists(file_path):
                self.logger.enable(log_level, base_level)
                self.fs.rm(file_path)
                self.logger.info(f"Fichier supprimé: '{file_path}'")
                return True
        except Exception as e:
            self.logger.enable(log_level, base_level)
            self.logger.error(f"Erreur lors de la suppression de '{file_path}': {e}")
            raise
        self.logger.enable(log_level, base_level)
        self.logger.warning(f"Tentative de suppression d'un fichier inexistant: '{file_path}'")
        return False

    @retry_on_failure
    @timer
    def create_empty_file(self, file_path: str) -> None:
        """Crée un fichier vide au chemin spécifié (ou tronque s'il existe)."""
        try:
            existed = self.fs.exists(file_path)
            with self.fs.open(file_path, mode="wb"):
                pass
            if existed:
                self.logger.info(f"Fichier tronqué: '{file_path}'")
            else:
                self.logger.info(f"Fichier vide créé: '{file_path}'")
        except Exception as e:
            self.logger.error(f"Erreur lors de la création du fichier vide '{file_path}': {e}")
            raise

    def _convertir_octets(self, taille_octets: int) -> tuple[float, str]:
        """Convertit une taille en octets dans une unité lisible."""
        for unite in ["octets", "Ko", "Mo", "Go", "To"]:
            if taille_octets < OCTETS_PAR_KO:
                return round(taille_octets, 2), unite # noqa: RUF057 indique que c'est un entier mais pas forcément
            taille_octets /= OCTETS_PAR_KO
        return round(taille_octets, 2), "To"

    def _ensure_table_exists(self, db: DataFactory, schema: str, table: str) -> None:
        """Crée la table cible si elle n'existe pas encore."""
        cols = ["tenantname", "date", "taille", "unite", "fichier"]
        cols_def = ["VARCHAR(255)", "TIMESTAMP", "FLOAT", "VARCHAR(10)", "VARCHAR(255)"]
        if isinstance(db, DbMssql):
            cols_def[1] = "DATETIME2"
        columns = ", ".join(f"{col} {typ}" for col, typ in zip(cols, cols_def))
        full_table = f"{schema}.{table}" if schema else table
        if isinstance(db, (DbMysql, DbPgsql)):
            query = f"CREATE TABLE IF NOT EXISTS {full_table} ({columns});"
        elif isinstance(db, DbMssql):
            query = f"""
            IF OBJECT_ID('{full_table}', 'U') IS NULL
            CREATE TABLE {full_table} ({columns});
            """
        else:
            msg = f"Type de base de données non pris en charge: {type(db).__name__}"
            raise TypeError(msg)
        db.sql_exec(query)

    def _collect_files(self, base_path: str, file_filter: str) -> list[str]:
        """Retourne la liste complète des fichiers correspondant au filtre."""
        files = []
        def _walk(path: str) -> None:
            self.logger.debug(f"[COLLECT] Exploration du dossier : {path}")
            try:
                entries = self.fs.ls(path)
                for entry in entries:
                    fp = entry["name"] if isinstance(entry, dict) else entry
                    self.logger.debug(f"[COLLECT] Entrée trouvée : {fp}")
                    if self.fs.isfile(fp):
                        if file_filter in fp:
                            files.append(fp)
                        else:
                            self.logger.debug(f"[COLLECT] Fichier ignoré: {fp}")
                    elif self.fs.isdir(fp):
                        self.logger.debug(f"[COLLECT] Répertoire à explorer : {fp}")
                        _walk(fp)
            except Exception as e:
                self.logger.error(f"Erreur lors de la collecte des fichiers dans {path}: {e}")
                raise
        _walk(base_path)
        return files

    def _process_files(
            self,
            db: DataFactory,
            schema: str,
            table: str,
            fichiers: list[str],
            tenant_name: str,
            date: str,
            base_path: str,
    ) -> None:
        """Traite les fichiers, calcule les tailles et insère les résultats dans la base."""
        total_size = 0
        bulk_data = []
        cols = ["tenantname", "date", "taille", "unite", "fichier"]
        try:
            for fichier in fichiers:
                info = self.fs.info(fichier)
                size = int(info.get("size", info.get("Size", 0)))
                total_size += size
                value, unit = self._convertir_octets(size)
                bulk_data.append([tenant_name, date, str(value), unit, fichier])
        except Exception as e:
            self.logger.error(f"Erreur sur le fichier '{fichier}': {e}")
            raise
        if bulk_data:
            db.insert_bulk(schema, table, cols, bulk_data)

        total_value, total_unit = self._convertir_octets(total_size)
        total_data = [f"{tenant_name} TOTAL", date, str(total_value), total_unit, f"TOTAL {base_path}"]
        db.insert(schema, table, cols, total_data)

        self.logger.info(f"Taille totale: {total_value} {total_unit}")

    @retry_on_failure
    @timer
    def disk_check(
            self,
           db: DataFactory,
           schema: str,
           table: str,
           base_path: str,
           tenant_name: str = "DEFAULT",
           file_filter: str = "python-",
    ) -> None:
        """
        Calcule la taille des fichiers correspondant à un filtre dans un répertoire et enregistre les résultats dans une base.

        :param db: Instance de data_factory (dbMssql, dbPgsql, dbMysql)
        :param schema: Nom du schéma
        :param table: Nom de la table cible
        :param base_path: Répertoire racine à scanner
        :param tenant_name: Nom du tenant à enregistrer
        :param file_filter: Mot-clé à rechercher dans les fichiers
        """  # noqa: D401
        self.logger.info(f"Lancement du disk_check à partir de {base_path}")
        db.connect()
        today = datetime.now(timezone.utc).date()
        self._ensure_table_exists(db, schema, table)
        fichiers = self._collect_files(base_path, file_filter)
        if not fichiers:
            self.logger.warning(f"Aucun fichier contenant '{file_filter}' trouvé dans '{base_path}'.")
            return
        self._process_files(db, schema, table, fichiers, tenant_name, str(today), base_path)

    def wait_for_file(self, path: str, filename: str, retry: int | None = None, delay: int | None = None) -> bool:
        """Attend qu'un fichier spécifique soit disponible dans un répertoire donné."""
        retry = retry or self.retry_count
        delay = delay or self.retry_delay
        self.logger.info(f"Attente du fichier '{filename}' dans le répertoire '{path}'.")
        found = False
        if not self.directory_exists(path):
            msg = f"Le dossier {path} ne semble pas exister."
            self.logger.error(msg)
            raise FileNotFoundError(msg)
        attempt = 0
        for attempt in range(1, retry + 1): # noqa: B007
            found = self.fs.exists(f"{path}/{filename}")
            if found:
                break
            time.sleep(delay)
        if found:
            self.logger.info(f"Fichier '{filename}' trouvé après {attempt} tentatives(s), soit {attempt * delay}s.")
        else:
            self.logger.warning(f"Fichier '{filename}' trouvé après {attempt} tentatives(s), soit {retry * delay}s.")
        return found
